import React, { useState, useEffect, useRef, useMemo } from 'react';
import * as d3 from 'd3';
import Pagination from './Pagination';
import BarPlot from './BarPlot';
import { NodeAttributes } from '@graphpolaris/shared/lib/data-access/store/graphQueryResultSlice';
import { SchemaAttributeTypes } from '@graphpolaris/shared/lib/schema';

export type AugmentedNodeAttributes = { attribute: NodeAttributes; type: Record<string, SchemaAttributeTypes> };

type TableProps = {
  data: AugmentedNodeAttributes[];
  itemsPerPage: number;
  showBarPlot: boolean;
};
type Data2RenderI = {
  name: string;
  typeData: SchemaAttributeTypes;
  data: { category: string; count: number }[];
  numElements: number;
};

export const Table = ({ data, itemsPerPage, showBarPlot }: TableProps) => {
  const maxUniqueValues = 100;
  const barPlotNumBins = 10;

  const [sortedData, setSortedData] = useState<AugmentedNodeAttributes[]>(data);
  const [sortOrder, setSortOrder] = useState<'asc' | 'desc'>('asc');
  const [sortColumn, setSortColumn] = useState<string | null>(null);
  const [currentPage, setCurrentPage] = useState<{
    page: number;
    startIndex: number;
    endIndex: number;
    currentData: AugmentedNodeAttributes[];
  } | null>(null);
  const [data2Render, setData2Render] = useState<Data2RenderI[]>([]);
  const dataColumns = useMemo(() => Object.keys(data[0].attribute), [data]);
  const totalPages = Math.ceil(sortedData.length / itemsPerPage);

  useEffect(() => {
    if (sortColumn !== null) {
      const sorted = [...data].sort((a, b) => {
        if (sortOrder === 'asc') {
          return (a.attribute as any)[sortColumn] < (b.attribute as any)[sortColumn] ? -1 : 1;
        } else {
          return (a.attribute as any)[sortColumn] > (b.attribute as any)[sortColumn] ? -1 : 1;
        }
      });
      setSortedData(sorted);
    }
  }, [sortOrder, data, sortColumn]);

  useEffect(() => {
    onPageChange(1); // Reset to the first page when sorting or itemsPerPage changes
  }, [sortColumn, sortOrder, itemsPerPage]);

  const onPageChange = (page: number) => {
    const startIndex = (page - 1) * itemsPerPage;
    const endIndex = startIndex + itemsPerPage;
    const currentData = sortedData.slice(startIndex, endIndex);

    setCurrentPage({
      page: page,
      startIndex: startIndex,
      endIndex: endIndex,
      currentData: currentData,
    });
  };

  const toggleSort = (column: string) => {
    if (sortColumn === column) {
      if (sortOrder === 'asc') {
        setSortOrder('desc');
        setSortedData(sortedData.reverse()); // Reverse the sorted data
      } else if (sortOrder === 'desc') {
        setSortColumn(null);
        setSortOrder('asc');
        setSortedData(data); // Reset to the original order
      }
    } else {
      setSortColumn(column);
      setSortOrder('asc');
      const sorted = [...data].sort((a, b) => {
        return (a.attribute as any)[column] < (b.attribute as any)[column] ? -1 : 1;
      });
      setSortedData(sorted);
    }
  };

  // Barplot on headers data preparation

  // Data structure to feed the keys-barplot ( name data2Render )
  // Three options:
  // 1) numeric->data to histogram barplot
  // 2) string with unique values < maxUniqueValues -> data binned barplot
  // 3) string with unique values > maxUniqueValues -> show text unique values

  useEffect(() => {
    if (!currentPage || currentPage?.currentData?.length <= 0) return;

    let categoryCounts = [];
    const firstRowData = data[0];

    let _data2Render = Object.keys(firstRowData.attribute).map((dataColumn: string, i) => {
      const newData2Render: Data2RenderI = {
        name: dataColumn,
        typeData: firstRowData.type[dataColumn] || 'string',
        data: [],
        numElements: 0,
      };

      if (firstRowData.type[dataColumn] === 'string') {
        const groupedData = d3.group(data, (d) => d.attribute[dataColumn]);

        categoryCounts = Array.from(groupedData, ([category, items]) => ({
          category: category as string,
          count: items.length,
        }));

        if (categoryCounts.length > maxUniqueValues) {
          newData2Render.numElements = categoryCounts.length;
        } else {
          newData2Render.numElements = categoryCounts.length;
          newData2Render.data = categoryCounts;
        }

        // number
        // perhaps check for other than string and number. eg. boolean
      } else {
        categoryCounts = data.map((obj) => ({
          category: 'placeholder', // add something
          count: obj.attribute[dataColumn] as number, // fill values of data
        }));

        // console.log('number', categoryCounts);
        newData2Render.numElements = categoryCounts.length;
        newData2Render.data = categoryCounts;
      }
      return newData2Render;
    });

    setData2Render(_data2Render);
  }, [currentPage, data, sortedData]);


  return (
    <div className="text-center font-inter text-primary">
      {currentPage && currentPage?.currentData?.length > 0 && data2Render?.length > 0 && (
        <>
          <table className="text-center my-2 mx-auto table-fixed w-11/12">
            <thead className="thead">
              <tr className="bg-white text-center p-0 pl-2 border-0 h-2 font-weight: 700">
                {dataColumns.map((item, index) => (
                  <th className="th cursor-pointer select-none" key={index + item} onClick={() => toggleSort(item)}>
                    {item} {sortColumn === item && <span>{sortOrder === 'asc' ? '▲' : '▼'}</span>}
                  </th>
                ))}
              </tr>
              <tr className="svggs align-top">
                {dataColumns.map((item, index) => (
                  <th className="th" key={index + item}>
                    <div className="h-full w-full overflow-hidden">
                      {showBarPlot && (data2Render[index]?.typeData === 'int' || data2Render[index]?.typeData === 'float') ? (
                        <BarPlot typeBarPlot='numerical' numBins={barPlotNumBins} data={data2Render[index].data} />
                      ) : !data2Render[index]?.typeData || !showBarPlot || data2Render[index].numElements > maxUniqueValues ? (
                        <div className="h-full text-xs font-normal flex flex-row items-center justify-center gap-1">
                          <span>Unique values:</span>
                          <span className="text-sm">{data2Render[index]?.numElements}</span>
                        </div>
                        ) : (
                        <BarPlot typeBarPlot='categorical' numBins={barPlotNumBins} data={data2Render[index].data} />
                      )}
                    </div>
                  </th>
                ))}
              </tr>
            </thead>
            <tbody className="border-l-2 border-t-2 border-r-2 border-b-2 border-white w-20">
              {currentPage.currentData.map((item, index) => (
                <tr key={index} className={index % 2 === 0 ? 'bg-base-200' : 'bg-base-100'}>
                  {dataColumns.map((col, index) => (
                    <td
                      className={`${item.type[col] === 'string' ? 'text-left' : 'text-center'
                      } px-1 py-0.5 border border-white m-0 overflow-x-hidden truncate`}
                      key={index}
                    >
                      {item.attribute[col] as string}
                    </td>
                  ))}
                </tr>
              ))}
            </tbody>
          </table>

          <Pagination
            currentPage={currentPage.page}
            totalPages={totalPages}
            onPageChange={onPageChange}
            itemsPerPageInput={itemsPerPage}
            numItemsArrayReal={currentPage.startIndex + currentPage.currentData.length}
            totalItems={sortedData.length}
          />
        </>
      )}
    </div>
  );
};

export default Table;
